@@ -2,42 +2,47 @@
     use strict;
     use warnings;
     use POSIX;
+    use Config;
     use Math::MPFR::Prec;
     use Math::MPFR::Random;
 
-    use constant  GMP_RNDN            => 0;
-    use constant  GMP_RNDZ            => 1;
-    use constant  GMP_RNDU            => 2;
-    use constant  GMP_RNDD            => 3;
-    use constant  MPFR_RNDN           => 0;
-    use constant  MPFR_RNDZ           => 1;
-    use constant  MPFR_RNDU           => 2;
-    use constant  MPFR_RNDD           => 3;
-    use constant  MPFR_RNDA           => 4;
-    use constant  _UOK_T              => 1;
-    use constant  _IOK_T              => 2;
-    use constant  _NOK_T              => 3;
-    use constant  _POK_T              => 4;
-    use constant  _MATH_MPFR_T        => 5;
-    use constant  _MATH_GMPf_T        => 6;
-    use constant  _MATH_GMPq_T        => 7;
-    use constant  _MATH_GMPz_T        => 8;
-    use constant  _MATH_GMP_T         => 9;
-    use constant  _MATH_MPC_T         => 10;
-    use constant MPFR_FLAGS_UNDERFLOW => 1;
-    use constant MPFR_FLAGS_OVERFLOW  => 2;
-    use constant MPFR_FLAGS_NAN       => 4;
-    use constant MPFR_FLAGS_INEXACT   => 8;
-    use constant MPFR_FLAGS_ERANGE    => 16;
-    use constant MPFR_FLAGS_DIVBY0    => 32;
-    use constant MPFR_FLAGS_ALL       => 63;
+    use constant  GMP_RNDN              => 0;
+    use constant  GMP_RNDZ              => 1;
+    use constant  GMP_RNDU              => 2;
+    use constant  GMP_RNDD              => 3;
+    use constant  MPFR_RNDN             => 0;
+    use constant  MPFR_RNDZ             => 1;
+    use constant  MPFR_RNDU             => 2;
+    use constant  MPFR_RNDD             => 3;
+    use constant  MPFR_RNDA             => 4;
+    use constant  MPFR_RNDF             => 5;
+    use constant  _UOK_T                => 1;
+    use constant  _IOK_T                => 2;
+    use constant  _NOK_T                => 3;
+    use constant  _POK_T                => 4;
+    use constant  _MATH_MPFR_T          => 5;
+    use constant  _MATH_GMPf_T          => 6;
+    use constant  _MATH_GMPq_T          => 7;
+    use constant  _MATH_GMPz_T          => 8;
+    use constant  _MATH_GMP_T           => 9;
+    use constant  _MATH_MPC_T           => 10;
+    use constant MPFR_FLAGS_UNDERFLOW   => 1;
+    use constant MPFR_FLAGS_OVERFLOW    => 2;
+    use constant MPFR_FLAGS_NAN         => 4;
+    use constant MPFR_FLAGS_INEXACT     => 8;
+    use constant MPFR_FLAGS_ERANGE      => 16;
+    use constant MPFR_FLAGS_DIVBY0      => 32;
+    use constant MPFR_FLAGS_ALL         => 63;
+    use constant MPFR_FREE_LOCAL_CACHE  => 1;
+    use constant MPFR_FREE_GLOBAL_CACHE => 2;
 
 
     use subs qw(MPFR_VERSION MPFR_VERSION_MAJOR MPFR_VERSION_MINOR
                 MPFR_VERSION_PATCHLEVEL MPFR_VERSION_STRING
                 RMPFR_PREC_MIN RMPFR_PREC_MAX
                 MPFR_DBL_DIG MPFR_LDBL_DIG MPFR_FLT128_DIG
-                GMP_LIMB_BITS GMP_NAIL_BITS);
+                GMP_LIMB_BITS GMP_NAIL_BITS
+                );
 
     use overload
     '++'   => \&overload_inc,
@@ -77,12 +82,13 @@
     require DynaLoader;
 
     @Math::MPFR::EXPORT_OK = qw(GMP_RNDN GMP_RNDZ GMP_RNDU GMP_RNDD
-MPFR_RNDN MPFR_RNDZ MPFR_RNDU MPFR_RNDD MPFR_RNDA
+MPFR_RNDN MPFR_RNDZ MPFR_RNDU MPFR_RNDD MPFR_RNDA MPFR_RNDF
 MPFR_FLAGS_UNDERFLOW MPFR_FLAGS_OVERFLOW MPFR_FLAGS_DIVBY0 MPFR_FLAGS_NAN
 MPFR_FLAGS_INEXACT MPFR_FLAGS_ERANGE MPFR_FLAGS_ALL
 MPFR_VERSION MPFR_VERSION_MAJOR MPFR_VERSION_MINOR
 MPFR_VERSION_PATCHLEVEL MPFR_VERSION_STRING RMPFR_VERSION_NUM
 RMPFR_PREC_MIN RMPFR_PREC_MAX
+MPFR_FREE_LOCAL_CACHE MPFR_FREE_GLOBAL_CACHE
 Rmpfr_randclear Rmpfr_randinit_mt
 Rmpfr_randinit_default Rmpfr_randinit_lc_2exp
 Rmpfr_randinit_lc_2exp_size Rmpfr_randseed Rmpfr_randseed_ui
@@ -104,7 +110,8 @@ Rmpfr_exp10 Rmpfr_exp2 Rmpfr_expm1 Rmpfr_fac_ui
 Rmpfr_fits_intmax_p Rmpfr_fits_sint_p Rmpfr_fits_slong_p
 Rmpfr_fits_sshort_p Rmpfr_fits_uint_p Rmpfr_fits_uintmax_p
 Rmpfr_fits_ulong_p Rmpfr_fits_IV_p Rmpfr_fits_UV_p
-Rmpfr_fits_ushort_p Rmpfr_floor Rmpfr_fma Rmpfr_frac Rmpfr_free_cache Rmpfr_gamma
+Rmpfr_fits_ushort_p Rmpfr_floor Rmpfr_fma Rmpfr_frac Rmpfr_gamma
+Rmpfr_free_cache Rmpfr_free_cache2 Rmpfr_free_pool
 Rmpfr_get_d Rmpfr_get_IV Rmpfr_get_UV Rmpfr_get_NV
 Rmpfr_get_d_2exp Rmpfr_get_d1 Rmpfr_get_default_prec Rmpfr_get_default_rounding_mode
 Rmpfr_get_emax Rmpfr_get_emax_max Rmpfr_get_emax_min Rmpfr_get_emin
@@ -134,7 +141,7 @@ TRmpfr_out_str
 Rmpfr_overflow_p Rmpfr_pow Rmpfr_pow_si Rmpfr_pow_ui Rmpfr_pow_z Rmpfr_prec_round
 Rmpfr_print_rnd_mode
 Rmpfr_random2 Rmpfr_reldiff Rmpfr_rint Rmpfr_rint_ceil
-Rmpfr_rint_floor Rmpfr_rint_round Rmpfr_rint_trunc Rmpfr_root Rmpfr_round
+Rmpfr_rint_floor Rmpfr_rint_round Rmpfr_rint_trunc Rmpfr_root Rmpfr_rootn_ui Rmpfr_round
 Rmpfr_sec Rmpfr_sech Rmpfr_set Rmpfr_set_d Rmpfr_set_default_prec
 Rmpfr_set_default_rounding_mode Rmpfr_set_emax Rmpfr_set_emin Rmpfr_set_erangeflag
 Rmpfr_set_exp Rmpfr_set_f Rmpfr_set_inexflag Rmpfr_set_inf Rmpfr_set_ld Rmpfr_set_LD
@@ -144,7 +151,7 @@ Rmpfr_set_prec_raw Rmpfr_set_q Rmpfr_set_si Rmpfr_set_si_2exp Rmpfr_set_sj
 Rmpfr_set_sj_2exp Rmpfr_set_str Rmpfr_set_ui Rmpfr_set_ui_2exp
 Rmpfr_set_uj Rmpfr_set_uj_2exp
 Rmpfr_set_DECIMAL64 Rmpfr_get_DECIMAL64 Rmpfr_set_float128 Rmpfr_get_float128
-Rmpfr_set_FLOAT128 Rmpfr_get_FLOAT128
+Rmpfr_set_FLOAT128 Rmpfr_get_FLOAT128 Rmpfr_set_DECIMAL128 Rmpfr_get_DECIMAL128
 Rmpfr_set_underflow Rmpfr_set_z Rmpfr_sgn Rmpfr_si_div Rmpfr_si_sub Rmpfr_sin
 Rmpfr_sin_cos Rmpfr_sinh_cosh
 Rmpfr_sinh Rmpfr_sqr Rmpfr_sqrt Rmpfr_sqrt_ui Rmpfr_strtofr Rmpfr_sub
@@ -170,21 +177,23 @@ Rmpfr_fmodquo Rmpfr_fpif_export Rmpfr_fpif_import Rmpfr_flags_clear Rmpfr_flags_
 Rmpfr_flags_test Rmpfr_flags_save Rmpfr_flags_restore Rmpfr_rint_roundeven Rmpfr_roundeven
 Rmpfr_nrandom Rmpfr_erandom Rmpfr_fmma Rmpfr_fmms Rmpfr_log_ui Rmpfr_gamma_inc Rmpfr_beta
 Rmpfr_round_nearest_away rndna
+atonv nvtoa atodouble Rmpfr_dot Rmpfr_get_str_ndigits
 );
 
-    our $VERSION = '3.36';
+    our $VERSION = '4.09_001';
     #$VERSION = eval $VERSION;
 
     DynaLoader::bootstrap Math::MPFR $VERSION;
 
     %Math::MPFR::EXPORT_TAGS =(mpfr => [qw(
 GMP_RNDN GMP_RNDZ GMP_RNDU GMP_RNDD
-MPFR_RNDN MPFR_RNDZ MPFR_RNDU MPFR_RNDD MPFR_RNDA
+MPFR_RNDN MPFR_RNDZ MPFR_RNDU MPFR_RNDD MPFR_RNDA MPFR_RNDF
 MPFR_FLAGS_UNDERFLOW MPFR_FLAGS_OVERFLOW MPFR_FLAGS_DIVBY0 MPFR_FLAGS_NAN
 MPFR_FLAGS_INEXACT MPFR_FLAGS_ERANGE MPFR_FLAGS_ALL
 MPFR_VERSION MPFR_VERSION_MAJOR MPFR_VERSION_MINOR
 MPFR_VERSION_PATCHLEVEL MPFR_VERSION_STRING RMPFR_VERSION_NUM
 RMPFR_PREC_MIN RMPFR_PREC_MAX
+MPFR_FREE_LOCAL_CACHE MPFR_FREE_GLOBAL_CACHE
 Rmpfr_randclear Rmpfr_randinit_mt
 Rmpfr_randinit_default Rmpfr_randinit_lc_2exp
 Rmpfr_randinit_lc_2exp_size Rmpfr_randseed Rmpfr_randseed_ui
@@ -206,7 +215,8 @@ Rmpfr_exp10 Rmpfr_exp2 Rmpfr_expm1 Rmpfr_fac_ui
 Rmpfr_fits_intmax_p Rmpfr_fits_sint_p Rmpfr_fits_slong_p
 Rmpfr_fits_sshort_p Rmpfr_fits_uint_p Rmpfr_fits_uintmax_p
 Rmpfr_fits_ulong_p Rmpfr_fits_IV_p Rmpfr_fits_UV_p
-Rmpfr_fits_ushort_p Rmpfr_floor Rmpfr_fma Rmpfr_frac Rmpfr_free_cache Rmpfr_gamma
+Rmpfr_fits_ushort_p Rmpfr_floor Rmpfr_fma Rmpfr_frac Rmpfr_gamma
+Rmpfr_free_cache Rmpfr_free_cache2 Rmpfr_free_pool
 Rmpfr_get_d Rmpfr_get_IV Rmpfr_get_UV Rmpfr_get_NV
 Rmpfr_get_d_2exp Rmpfr_get_d1 Rmpfr_get_default_prec Rmpfr_get_default_rounding_mode
 Rmpfr_get_emax Rmpfr_get_emax_max Rmpfr_get_emax_min Rmpfr_get_emin
@@ -236,7 +246,7 @@ TRmpfr_out_str
 Rmpfr_overflow_p Rmpfr_pow Rmpfr_pow_si Rmpfr_pow_ui Rmpfr_pow_z Rmpfr_prec_round
 Rmpfr_print_rnd_mode
 Rmpfr_random2 Rmpfr_reldiff Rmpfr_rint Rmpfr_rint_ceil
-Rmpfr_rint_floor Rmpfr_rint_round Rmpfr_rint_trunc Rmpfr_root Rmpfr_round
+Rmpfr_rint_floor Rmpfr_rint_round Rmpfr_rint_trunc Rmpfr_root Rmpfr_rootn_ui Rmpfr_round
 Rmpfr_sec Rmpfr_sech Rmpfr_set Rmpfr_set_d Rmpfr_set_default_prec
 Rmpfr_set_default_rounding_mode Rmpfr_set_emax Rmpfr_set_emin Rmpfr_set_erangeflag
 Rmpfr_set_exp Rmpfr_set_f Rmpfr_set_inexflag Rmpfr_set_inf Rmpfr_set_ld Rmpfr_set_LD
@@ -246,7 +256,7 @@ Rmpfr_set_prec_raw Rmpfr_set_q Rmpfr_set_si Rmpfr_set_si_2exp Rmpfr_set_sj
 Rmpfr_set_sj_2exp Rmpfr_set_str Rmpfr_set_ui Rmpfr_set_ui_2exp
 Rmpfr_set_uj Rmpfr_set_uj_2exp
 Rmpfr_set_DECIMAL64 Rmpfr_get_DECIMAL64 Rmpfr_set_float128 Rmpfr_get_float128
-Rmpfr_set_FLOAT128 Rmpfr_get_FLOAT128
+Rmpfr_set_FLOAT128 Rmpfr_get_FLOAT128 Rmpfr_set_DECIMAL128 Rmpfr_get_DECIMAL128
 Rmpfr_set_underflow Rmpfr_set_z Rmpfr_sgn Rmpfr_si_div Rmpfr_si_sub Rmpfr_sin
 Rmpfr_sin_cos Rmpfr_sinh_cosh
 Rmpfr_sinh Rmpfr_sqr Rmpfr_sqrt Rmpfr_sqrt_ui Rmpfr_strtofr Rmpfr_sub
@@ -272,14 +282,18 @@ Rmpfr_fmodquo Rmpfr_fpif_export Rmpfr_fpif_import Rmpfr_flags_clear Rmpfr_flags_
 Rmpfr_flags_test Rmpfr_flags_save Rmpfr_flags_restore Rmpfr_rint_roundeven Rmpfr_roundeven
 Rmpfr_nrandom Rmpfr_erandom Rmpfr_fmma Rmpfr_fmms Rmpfr_log_ui Rmpfr_gamma_inc Rmpfr_beta
 Rmpfr_round_nearest_away rndna
+atonv nvtoa atodouble Rmpfr_dot Rmpfr_get_str_ndigits
 )]);
 
+
 $Math::MPFR::NNW = 0; # Set to 1 to allow "non-numeric" warnings for operations involving
                       # strings that contain non-numeric characters.
 
 $Math::MPFR::NOK_POK = 0; # Set to 1 to allow warnings in new() and overloaded operations when
                           # a scalar that has set both NOK (NV) and POK (PV) flags is encountered
 
+%Math::MPFR::NV_properties = _get_NV_properties();
+
 sub dl_load_flags {0} # Prevent DynaLoader from complaining and croaking
 
 sub Rmpfr_out_str {
@@ -388,7 +402,7 @@ sub new {
       if(!@_) {return Rmpfr_init()}
       }
 
-    # @_ can now contain a maximum of 2 args - the value, and iff the value is
+    # @_ can now contain a maximum of 2 args - the value, and if the value is
     # a string, (optionally) the base of the numeric string.
     if(@_ > 2) {die "Too many arguments supplied to new() - expected no more than two"}
 
@@ -545,16 +559,16 @@ sub Rmpfr_inits2_nobless {
     return @ret;
 }
 
-sub MPFR_VERSION {return _MPFR_VERSION()}
-sub MPFR_VERSION_MAJOR {return _MPFR_VERSION_MAJOR()}
-sub MPFR_VERSION_MINOR {return _MPFR_VERSION_MINOR()}
-sub MPFR_VERSION_PATCHLEVEL {return _MPFR_VERSION_PATCHLEVEL()}
-sub MPFR_VERSION_STRING {return _MPFR_VERSION_STRING()}
-sub MPFR_DBL_DIG {return _DBL_DIG()}
-sub MPFR_LDBL_DIG {return _LDBL_DIG()}
-sub MPFR_FLT128_DIG {return _FLT128_DIG()}
-sub GMP_LIMB_BITS {return _GMP_LIMB_BITS()}
-sub GMP_NAIL_BITS {return _GMP_NAIL_BITS()}
+sub MPFR_VERSION            () {return _MPFR_VERSION()}
+sub MPFR_VERSION_MAJOR      () {return _MPFR_VERSION_MAJOR()}
+sub MPFR_VERSION_MINOR      () {return _MPFR_VERSION_MINOR()}
+sub MPFR_VERSION_PATCHLEVEL () {return _MPFR_VERSION_PATCHLEVEL()}
+sub MPFR_VERSION_STRING     () {return _MPFR_VERSION_STRING()}
+sub MPFR_DBL_DIG            () {return _DBL_DIG()}
+sub MPFR_LDBL_DIG           () {return _LDBL_DIG()}
+sub MPFR_FLT128_DIG         () {return _FLT128_DIG()}
+sub GMP_LIMB_BITS           () {return _GMP_LIMB_BITS()}
+sub GMP_NAIL_BITS           () {return _GMP_NAIL_BITS()}
 
 sub mpfr_min_inter_prec {
     die "Wrong number of args to minimum_intermediate_prec()" if @_ != 3;
@@ -720,6 +734,67 @@ sub Rmpfr_round_nearest_away {
   return $ret;
 }
 
+sub _get_NV_properties {
+
+  my($bits, $PREC, $max_dig, $min_pow, $normal_min, $NV_MAX, $nvtype, $emax, $emin);
+
+  if   ($Config{nvtype} eq 'double')     {
+    $bits = 53;  $PREC = 64;  $max_dig = 17; $min_pow = -1074;
+    $normal_min = 2 ** -1022; $NV_MAX = POSIX::DBL_MAX; $emin = -1073; $emax = 1024;
+  }
+
+  elsif($Config{nvtype} eq '__float128') {
+    $bits = 113; $PREC = 128; $max_dig = 36; $min_pow = -16494; $normal_min = 2 ** -16382;
+    $NV_MAX = 1.18973149535723176508575932662800702e+4932; $emin = -16493; $emax = 16384;
+  }
+
+  elsif($Config{nvtype} eq 'long double') {
+
+    if(_required_ldbl_mant_dig() == 53)      {
+      $bits = 53;  $PREC = 64;  $max_dig = 17; $min_pow = -1074;
+      $normal_min = 2 ** -1022; $NV_MAX = POSIX::DBL_MAX; $emin = -1073; $emax = 1024;
+    }
+
+    elsif(_required_ldbl_mant_dig() == 113)  {
+      $bits = 113; $PREC = 128; $max_dig = 36; $min_pow = -16494;
+      $normal_min = 2 ** -16382; $NV_MAX = POSIX::LDBL_MAX; $emin = -16493; $emax = 16384;
+    }
+
+    elsif(_required_ldbl_mant_dig() == 64)   {
+      $bits = 64;  $PREC = 80;  $max_dig = 21; $min_pow = -16445;
+      $normal_min = 2 ** -16382; $NV_MAX = POSIX::LDBL_MAX; $emin = -16444; $emax = 16384;
+    }
+
+    elsif(_required_ldbl_mant_dig() == 2098) {
+      $bits = 2098;  $PREC = 2104;  $max_dig = 33; $min_pow = -1074;
+      $normal_min = 2 ** -1022; $NV_MAX = POSIX::LDBL_MAX; $emin = -1073; $emax = 1024;
+    }
+
+    else {
+      my %properties = ('type' => 'unknown long double type');
+      return %properties;
+    }
+  }
+  else {
+      my %properties = ('type' => 'unknown nv type');
+      return %properties;
+  }
+
+  my %properties = (
+    'bits'       => $bits,
+    'PREC'       => $PREC,
+    'max_dig'    => $max_dig,
+    'min_pow'    => $min_pow,
+    'normal_min' => $normal_min,
+    'NV_MAX'     => $NV_MAX,
+    'emin'       => $emin,
+    'emax'       => $emax,
+                   );
+
+  return %properties;
+}
+
+
 *Rmpfr_get_z_exp             = \&Rmpfr_get_z_2exp;
 *prec_cast                   = \&Math::MPFR::Prec::prec_cast;
 *Rmpfr_randinit_default      = \&Math::MPFR::Random::Rmpfr_randinit_default;
